import logging
from cisco_telescope.instrumentations.requests import RequestsInstrumentorWrapper
from cisco_telescope.instrumentations.aiohttp import AiohttpInstrumentorWrapper
from cisco_telescope.instrumentations.grpc import GrpcInstrumentorServerWrapper
from cisco_telescope.instrumentations.grpc import GrpcInstrumentorClientWrapper
from .. import consts, options


class InstrumentationWrapper:

    _INSTRUMENTATION_STATE = {}

    @classmethod
    def get_wrappers(cls):
        return {
            consts.REQUESTS_KEY: RequestsInstrumentorWrapper,
            consts.AIOHTTP_KEY: AiohttpInstrumentorWrapper,
            consts.GRPC_SERVER_KEY: GrpcInstrumentorServerWrapper,
            consts.GRPC_CLIENT_KEY: GrpcInstrumentorClientWrapper,
        }

    @classmethod
    def is_already_instrumented(cls, library_key):
        """check if an instrumentation wrapper is already registered"""
        return library_key in cls._INSTRUMENTATION_STATE.keys()

    @classmethod
    def get_instrumentation_wrapper(cls, opt: options.Options, library_key):
        """load and initialize an instrumentation wrapper"""
        if cls.is_already_instrumented(library_key):
            return None
        try:
            inst_dict = InstrumentationWrapper.get_wrappers()
            if library_key in inst_dict:
                wrapper_object = inst_dict[library_key]
                wrapper_instance = wrapper_object()
                cls._mark_as_instrumented(library_key, wrapper_instance)
                return wrapper_instance
            else:
                logging.info(f"No instrumentation wrapper for {library_key}")
                return
        except Exception:
            logging.exception(
                f"Error while attempting to load instrumentation wrapper for {library_key}"
            )
            return

    @classmethod
    def uninstrument_all(cls):
        for key in cls._INSTRUMENTATION_STATE:
            cls._INSTRUMENTATION_STATE[key].uninstrument()

        cls._INSTRUMENTATION_STATE.clear()

    @classmethod
    def _mark_as_instrumented(cls, library_key, wrapper_instance):
        """mark an instrumentation wrapper as registered"""
        cls._INSTRUMENTATION_STATE[library_key] = wrapper_instance



"""
Copyright The Cisco Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import asyncio
import types
from collections import deque

import aiohttp
import wrapt

from opentelemetry.context import create_key
from opentelemetry import context as context_api
from opentelemetry.instrumentation.aiohttp_client import (
    AioHttpClientInstrumentor,
    _ResponseHookT,
    _RequestHookT,
)
from opentelemetry.instrumentation.aiohttp_client.version import __version__
from opentelemetry.instrumentation.utils import http_status_to_status_code
from opentelemetry.propagate import inject
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import (
    Span,
    TracerProvider,
    get_tracer,
    Status,
    SpanKind,
    StatusCode,
)
from opentelemetry import trace
from cisco_opentelemetry_specifications import SemanticAttributes

from ..utils import Utils
from ... import consts


def request_hook(span: Span, params: aiohttp.TraceRequestStartParams) -> None:
    if not span or not span.is_recording():
        return

    Utils.add_flattened_dict(
        span,
        SemanticAttributes.HTTP_REQUEST_HEADER,
        getattr(params, "headers", dict()),
    )


def response_hook(
    span: Span,
    params: aiohttp.TraceRequestEndParams,
) -> None:
    if not span or not span.is_recording():
        return

    if hasattr(params, "response"):
        Utils.add_flattened_dict(
            span,
            SemanticAttributes.HTTP_RESPONSE_HEADER,
            getattr(params.response, "headers", dict()),
        )


class AiohttpInstrumentorWrapper(AioHttpClientInstrumentor):
    def __init__(self):
        super().__init__()

    def _instrument(self, **kwargs) -> None:
        _instrument(
            tracer_provider=kwargs.get("tracer_provider"),
            request_hook=request_hook,
            response_hook=response_hook,
        )

    def _uninstrument(self, **kwargs) -> None:
        super()._uninstrument()


def create_trace_config(
    tracer_provider: TracerProvider = None,
    request_hook: _RequestHookT = None,
    response_hook: _ResponseHookT = None,
) -> aiohttp.TraceConfig:
    tracer = get_tracer(__name__, __version__, tracer_provider)

    def _end_trace(trace_config_ctx: types.SimpleNamespace):
        context_api.detach(trace_config_ctx.token)
        trace_config_ctx.span.end()

    async def on_request_start(
        unused_session: aiohttp.ClientSession,
        trace_config_ctx: types.SimpleNamespace,
        params: aiohttp.TraceRequestStartParams,
    ):
        if context_api.get_value(create_key("suppress_instrumentation")):
            trace_config_ctx.span = None
            return

        http_method = params.method.upper()
        request_span_name = f"HTTP {http_method}"

        trace_config_ctx.span = trace_config_ctx.tracer.start_span(
            request_span_name,
            kind=SpanKind.CLIENT,
        )

        if callable(request_hook):
            request_hook(trace_config_ctx.span, params)

        if trace_config_ctx.span.is_recording():
            attributes = {
                SpanAttributes.HTTP_METHOD: http_method,
            }
            for key, value in attributes.items():
                trace_config_ctx.span.set_attribute(key, value)

        trace_config_ctx.token = context_api.attach(
            trace.set_span_in_context(trace_config_ctx.span)
        )

        inject(params.headers)

    async def on_request_end(
        unused_session: aiohttp.ClientSession,
        trace_config_ctx: types.SimpleNamespace,
        params: aiohttp.TraceRequestEndParams,
    ):
        if trace_config_ctx.span is None:
            return

        if callable(response_hook):
            response_hook(trace_config_ctx.span, params)

        if trace_config_ctx.span.is_recording():
            trace_config_ctx.span.set_status(
                Status(http_status_to_status_code(int(params.response.status)))
            )
            trace_config_ctx.span.set_attribute(
                SpanAttributes.HTTP_STATUS_CODE, params.response.status
            )
            # Expand opentelemetry implementations
            handle_request_body(trace_config_ctx)
            await handle_response_body(trace_config_ctx, params)

        _end_trace(trace_config_ctx)

    async def on_request_exception(
        unused_session: aiohttp.ClientSession,
        trace_config_ctx: types.SimpleNamespace,
        params: aiohttp.TraceRequestExceptionParams,
    ):
        if trace_config_ctx.span is None:
            return

        if callable(response_hook):
            response_hook(trace_config_ctx.span, params)

        if trace_config_ctx.span.is_recording() and params.exception:
            trace_config_ctx.span.set_status(Status(StatusCode.ERROR))
            trace_config_ctx.span.record_exception(params.exception)
        _end_trace(trace_config_ctx)

    async def on_request_chunk_sent(
        unused_session: aiohttp.ClientSession,
        trace_config_ctx: types.SimpleNamespace,
        params: aiohttp.TraceRequestChunkSentParams,
    ):
        if getattr(params, "chunk"):
            decoded_chunk = params.chunk.decode()
            trace_config_ctx.request_body += decoded_chunk

    async def on_responde_chunk_received(
        unused_session: aiohttp.ClientSession,
        trace_config_ctx: types.SimpleNamespace,
        params: aiohttp.TraceRequestChunkSentParams,
    ):
        if getattr(params, "chunk"):
            decoded_chunk = params.chunk.decode()
            trace_config_ctx.response_body += decoded_chunk

    def _trace_config_ctx_factory(**kwargs):
        kwargs.setdefault("trace_request_ctx", {})
        return types.SimpleNamespace(
            tracer=tracer, **kwargs, request_body="", response_body=""
        )

    trace_config = aiohttp.TraceConfig(
        trace_config_ctx_factory=_trace_config_ctx_factory
    )

    trace_config.on_request_start.append(on_request_start)
    trace_config.on_request_chunk_sent.append(on_request_chunk_sent)
    trace_config.on_response_chunk_received.append(on_responde_chunk_received)
    trace_config.on_request_exception.append(on_request_exception)
    trace_config.on_request_end.append(on_request_end)

    return trace_config


def _instrument(
    tracer_provider: TracerProvider = None,
    request_hook: _RequestHookT = None,
    response_hook: _ResponseHookT = None,
):
    def instrumented_init(wrapped, instance, args, kwargs) -> None:
        trace_configs = list(kwargs.get("trace_configs") or ())

        trace_config = create_trace_config(
            tracer_provider=tracer_provider,
            request_hook=request_hook,
            response_hook=response_hook,
        )

        trace_configs.append(trace_config)

        kwargs["trace_configs"] = trace_configs
        return wrapped(*args, **kwargs)

    wrapt.wrap_function_wrapper(aiohttp.ClientSession, "__init__", instrumented_init)


def handle_request_body(trace_config_ctx: types.SimpleNamespace):
    if getattr(trace_config_ctx, "request_body"):
        Utils.set_payload(
            trace_config_ctx.span,
            SemanticAttributes.HTTP_REQUEST_BODY,
            trace_config_ctx.request_body,
        )


async def handle_response_body(
    trace_config_ctx: types.SimpleNamespace, params: aiohttp.TraceRequestEndParams
):
    response_body = b""
    if getattr(params.response, "content"):
        # copy response content into temporary queue
        content_stream = params.response.content
        tmp_deque = deque()
        try:
            while not content_stream.at_eof():
                response_chunk = b""  # verify var has value
                response_chunk = await asyncio.wait_for(
                    content_stream.read(), consts.MAX_WAIT_TIME
                )
                tmp_deque.append(response_chunk)
                response_body += response_chunk

        finally:
            # retrieve response data from temporary queue
            content_stream._cursor = 0

            content_stream._buffer = tmp_deque

            Utils.set_payload(
                trace_config_ctx.span,
                SemanticAttributes.HTTP_RESPONSE_BODY,
                response_body,
            )



from typing import AnyStr
from opentelemetry.trace.span import Span
from cisco_opentelemetry_specifications.payload_attributes import PAYLOAD_ATTRIBUTES
from cisco_telescope.configuration import Configuration

from .. import consts


class Utils(object):
    @staticmethod
    def set_payload(
        span: Span,
        attr: str,
        payload: AnyStr,
    ):
        payload_decoded = payload or ""
        payloads_enabled = Configuration().payloads_enabled
        max_payload_size = Configuration().max_payload_size

        if payloads_enabled or (attr not in PAYLOAD_ATTRIBUTES):
            if isinstance(payload, bytes):
                payload_decoded = payload.decode(
                    consts.ENCODING_UTF8, consts.DECODE_PAYLOAD_IN_CASE_OF_ERROR
                )

            span.set_attribute(attr, payload_decoded[:max_payload_size])

    @staticmethod
    def lowercase_items(items: dict):
        return {k.lower(): v for k, v in items.items()}

    @staticmethod
    def add_flattened_dict(span: Span, prefix, attributes: dict):
        """Add Dictionary to Span as flattened labels with lower cased key values"""
        payloads_enabled = Configuration().payloads_enabled
        if payloads_enabled or (prefix not in PAYLOAD_ATTRIBUTES):
            for attribute_key, attribute_value in Utils.lowercase_items(
                attributes
            ).items():
                span.set_attribute(f"{prefix}.{attribute_key}", attribute_value)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cisco_telescope',
 'cisco_telescope.cmd',
 'cisco_telescope.instrumentations',
 'cisco_telescope.instrumentations.aiohttp',
 'cisco_telescope.instrumentations.grpc',
 'cisco_telescope.instrumentations.requests']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'argparse>=1.4.0,<2.0.0',
 'cisco-opentelemetry-specifications>=0.0.15,<0.0.16',
 'importlib-metadata>=4.11.3,<5.0.0',
 'opentelemetry-exporter-otlp>=1.10.0,<2.0.0',
 'opentelemetry-instrumentation-aiohttp-client>=0.29b0,<0.30',
 'opentelemetry-instrumentation-grpc==0.29b0',
 'opentelemetry-instrumentation-requests>=0.29b0,<0.30',
 'opentelemetry-instrumentation>=0.29b0,<0.30',
 'opentelemetry-sdk>=1.10.0,<2.0.0',
 'opentelemetry-semantic-conventions>=0.29b0,<0.30',
 'requests>=2.27.1,<3.0.0']

entry_points = \
{'console_scripts': ['telescope = cisco_telescope.cmd.trace:run',
                     'telescope-bootstrap = cisco_telescope.cmd.bootstrap:run']}

setup_kwargs = {
    'name': 'cisco-telescope',
    'version': '1.1.1',
    'description': 'Cisco Distribution for OpenTelemetry',
    'long_description': '# otel-py\n[![PIP Published Version][pip-image]][pip-url]\n[![Apache License][license-image]][license-image]\n[![Coverage][coverage-image]][coverage-url]\n\n![Trace](trace.png)\n\nThis package provides OpenTelemetry-compliant tracing to Python\napplications for the collection of distributed tracing and performance metrics in [Cisco Telescope](https://console.telescope.app/?utm_source=github).\n\n## Contents\n\n- [Installation](#installation)\n  - [Install Packages](#install-packages)\n  - [Install Dependencies](#install-dependencies)\n  - [Library initialization](#library-initialization)\n  - [OpenTelemetry Collector Configuration](#opentelemetry-collector-configuration)\n  - [Existing OpenTelemetry Instrumentation](#existing-opentelemetry-instrumentation)\n- [Supported Runtimes](#supported-runtimes)\n- [Supported Libraries](#supported-libraries)\n- [Configuration](#configuration)\n- [Getting Help](#getting-help)\n- [Opening Issues](#opening-issues)\n- [License](#license)\n\n## Installation\n\n### Install packages\n> To install Cisco OpenTelemetry Distribution simply run:\n\n```sh\npip install cisco-telescope\n```\n\n### Install dependencies\nTo install all supported instrumentation frameworks run:\n```sh\ntelescope-bootstrap\n```\n\n\n### Library initialization\nCisco OpenTelemetry Distribution is activated and instruments the supported libraries once the `ciscoTracing.init()` has returned.\n\nTo initialize the library, you\'ll need a cisco-token, which is taken from your [Account tab on the Telescope console Settings page](https://console.telescope.app/settings/account).\n\n\n```python\nfrom cisco_telescope import tracing\n\ntracing.init(\n  service_name="<your-service-name>",\n  cisco_token="<your-cisco-token>"\n)\n```\n\n### OpenTelemetry Collector Configuration\n\nBy default, Cisco OpenTelemetry Distribution exports data directly to [Cisco Telescope\'s](https://console.telescope.app/?utm_source=github) infrastructure backend.\n**Existing** OpenTelemetery Collector is supported, the following configuration can be applied\n\n#### Configure custom trace exporter\n\nCisco OpenTelemetry Distribution supports the configuration of multiple custom exporters. Note that you will need to handle your exporter authorization.\nExample for creating OtlpGrpc Span exporter to local OpenTelemetry collector:\n\n```python\nfrom cisco_telescope import tracing, options\n\ntracing.init(\n  service_name="<your-service-name>",\n  exporters=[\n    options.ExporterOptions(\n      exporter_type="otlp-grpc",\n      collector_endpoint="grpc://localhost:4317"\n    ),\n  ]\n)\n\n```\n\n#### Configure custom OpenTelemetry collector to export trace data to [Cisco Telescope\'s](https://console.telescope.app/?utm_source=github) external collector.\n\n```yaml\ncollector.yaml ...\n\nexporters:\n  otlphttp:\n    traces_endpoint: https://production.cisco-udp.com/trace-collector\n    headers:\n      authorization: Bearer <Your Telescope Token>\n    compression: gzip\n\n\nservice:\n  pipelines:\n    traces:\n      exporters: [otlphttp]\n```\n\n### Existing OpenTelemetry Instrumentation\n\n> Notice: Only relevant if interested in streaming existing OpenTelemetry workloads.\n> [Cisco Telescope](https://console.telescope.app/?utm_source=github). supports native OpenTelemetery traces.\n```python\nfrom opentelemetry import trace\nfrom opentelemetry.sdk.resources import Resource\nfrom opentelemetry.sdk.trace import TracerProvider\nfrom opentelemetry.sdk.trace.export import BatchSpanProcessor\n\nfrom opentelemetry.exporter.otlp.proto.http.trace_exporter import (\n    OTLPSpanExporter as OTLPHTTPExporter,\n)\n\nprovider = TracerProvider(resource=Resource.create())\ntrace.set_tracer_provider(provider)\n\nhttp_exporter = OTLPHTTPExporter(\n  endpoint="https://production.cisco-udp.com/trace-collector",\n  headers= {\n    "authorization": "Bearer <Your Telescope Token>",\n  },\n)\n\nprocessor = BatchSpanProcessor(http_exporter)\nprovider.add_span_processor(processor)\n```\n\n## Supported Runtimes\nCisco OpenTelemetry Distribution supports Python 3.6+\n\n## Supported Libraries\n\n> Cisco OpenTelemetry Python Distribution is extending Native OpenTelemetry, supported libraries [available here](https://github.com/open-telemetry/opentelemetry-python-contrib/tree/main/instrumentation).\n\nCisco OpenTelemetry Python Distribution provides out-of-the-box instrumentation (tracing) and advanced **payload collections** for many popular frameworks and libraries.\n\n| Library  | Extended Support Version    |\n|----------|-----------------------------|\n| requests | Fully supported             |\n| aiohttp  | Fully supported             |\n| grpc     | Client/Server Unary support |\n\n## Configuration\n\nAdvanced options can be configured as a parameter to the init() method:\n\n| Parameter        | Env                    | Type    | Default       | Description                                                       |\n|------------------|------------------------| ------- |---------------| ----------------------------------------------------------------- |\n| cisco_token      | CISCO_TOKEN            | string  | -             | Cisco account token                                               |\n| service_name     | OTEL_SERVICE_NAME      | string  | `application` | Application name that will be set for traces                      |\n| debug            | CISCO_DEBUG            | string  | `False`       | Debug logs                                                        |\n| payloads_enabled | CISCO_PAYLOADS_ENABLED | string  | `True`        | Debug logs                                                        |\n\nExporter options\n\n| Parameter          | Env                     | Type                | Default                                               | Description                                                                                                                                |\n|--------------------| ----------------------- | ------------------- | ----------------------------------------------------- |--------------------------------------------------------------------------------------------------------------------------------------------|\n| collector_endpoint | OTEL_COLLECTOR_ENDPOINT | string              | `https://production.cisco-udp.com/trace-collector` | The address of the trace collector to send traces to                                                                                       |\n| type               | OTEL_EXPORTER_TYPE      | string              | `otlp-http`                                           | The exporter type to use (Currently only `otlp-http` is supported). Multiple exporter option available via init function see example below |\n\n## Getting Help\n\nIf you have any issue around using the library or the product, please don\'t hesitate to:\n\n- Use the [documentation](https://docs.telescope.app).\n- Use the help widget inside the product.\n- Open an issue in GitHub.\n\n## License\n\nProvided under the Apache 2.0. See LICENSE for details.\n\nCopyright 2022, Cisco\n\n[pip-url]: https://pypi.org/project/cisco-telescope/\n[pip-image]: https://img.shields.io/github/v/release/cisco-open/otel-py?include_prereleases&style=for-the-badge\n[license-url]: https://github.com/https://github.com/cisco-open/otel-py/blob/main/LICENSE\n[license-image]: https://img.shields.io/badge/license-Apache_2.0-green.svg?style=for-the-badge\n[coverage-url]: https://codecov.io/gh/cisco-open/otel-py/branch/main/\n[coverage-image]: https://img.shields.io/codecov/c/github/cisco-open/otel-py?style=for-the-badge\n',
    'author': 'Cisco Epsagon Team',
    'author_email': 'support@epsagon.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cisco-open/otel-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from . import options

from cisco_opentelemetry_specifications.consts import Consts


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


@singleton
class Configuration(object):
    def __init__(self):
        self.reset_to_default()

    def reset_to_default(self):
        self._max_payload_size: int = Consts.DEFAULT_MAX_PAYLOAD_SIZE
        self._payloads_enabled: bool = Consts.DEFAULT_PAYLOADS_ENABLED

    def set_options(self, opt: options.Options):
        self._max_payload_size = opt.max_payload_size
        self._payloads_enabled = opt.payloads_enabled

    @property
    def max_payload_size(self):
        return self._max_payload_size

    @max_payload_size.setter
    def max_payload_size(self, value):
        self._max_payload_size = value

    @property
    def payloads_enabled(self):
        return self._payloads_enabled

    @payloads_enabled.setter
    def payloads_enabled(self, value):
        self._payloads_enabled = value
